@@ -5,7 +5,7 @@
 #[macro_use]
 extern crate build_cfg;
 
-use std::{env, path::PathBuf};
+use std::{env, io, path::{Path, PathBuf}};
 
 /// The error type returned by [`main`].
 #[derive(Debug)]
@@ -14,32 +14,36 @@ pub enum Error {
     SyncDep(anyhow::Error),
     /// Failed to build *libui*.
     #[cfg(feature = "build")]
-    BuildLibui(libui::Error),
-    /// Failed to generate bindings.
+    BuildLibui(build::Error),
+    /// Failed to include Windows resources.
+    IncludeWinres(io::Error),
+    /// Failed to generate bindings to *libui*.
     GenBindings(bindings::Error),
 }
 
 #[build_cfg_main]
 fn main() -> Result<(), Error> {
     let out_dir = PathBuf::from(env::var("OUT_DIR").unwrap());
-
     let libui_dir = out_dir.join("libui-ng");
     let meson_dir = out_dir.join("meson");
     let ninja_dir = out_dir.join("ninja");
 
-    // Cargo will prevent this crate from being published if the build script modifies files
-    // outside `$OUT_DIR` during its operation. To work around this for the purpose of building
-    // *libui*, we copy all non-Rust build dependencies to `$OUT_DIR`.
+    // Cargo will prevent this crate from being published if the build script modifies files outside
+    // `$OUT_DIR` during its operation. To work around this for the purpose of building *libui*, we
+    // copy all non-Rust build dependencies to `$OUT_DIR`.
     dep::sync("libui-ng", &libui_dir).map_err(Error::SyncDep)?;
-    dep::sync("meson", &meson_dir).map_err(Error::SyncDep)?;
-
-    if build_cfg!(target_os = "linux") {
-        dep::sync("ninja", &ninja_dir).map_err(Error::SyncDep)?;
-    }
 
     #[cfg(feature = "build")]
     {
-        libui::build(&libui_dir, &meson_dir, &ninja_dir).map_err(Error::BuildLibui)?;
+        let backend = build::Backend::default();
+
+        dep::sync("meson", &meson_dir).map_err(Error::SyncDep)?;
+        // Ninja only needs to be synced if it's selected as a build backend.
+        if let build::Backend::Ninja = backend {
+            dep::sync("ninja", &ninja_dir).map_err(Error::SyncDep)?;
+        }
+
+        backend.build_libui(&libui_dir, &meson_dir, &ninja_dir).map_err(Error::BuildLibui)?;
 
         // Tell Cargo where to find the copy of *libui* that we just built.
         println!(
@@ -52,17 +56,14 @@ fn main() -> Result<(), Error> {
         // shared objects that must be imported, we must tell Cargo (and, by extension, the dynamic
         // linker) which shared objects we need.
         import_dylibs();
+
+        if build_cfg!(target_os = "windows") {
+            include_winres(&libui_dir).map_err(Error::IncludeWinres)?;
+        }
     }
 
     // Instruct Cargo to link to *libui*.
-    println!(
-        "cargo:rustc-link-lib={}=ui",
-        if cfg!(feature = "build") {
-            "static"
-        } else {
-            "dylib"
-        },
-    );
+    println!("cargo:rustc-link-lib={}=ui", link_kind());
 
     bindings::generate(&libui_dir, &out_dir).map_err(Error::GenBindings)?;
 
@@ -84,7 +85,7 @@ fn import_dylibs() {
 
     if build_cfg!(target_os = "linux") {
         // While unintuitive, we don't actually need to specify any shared objects here---the
-        // `pkg_config` crate will do that automatically in [`bindings::ClangArgs::new_unix`].
+        // `pkg_config` crate will do that automatically in [`bindings::ClangArgs::new_linux`].
     } else if build_cfg!(target_os = "windows") {
         // See `dep/libui-ng/windows/meson.build`.
         dyn_link! {
@@ -106,6 +107,21 @@ fn import_dylibs() {
     }
 }
 
+fn include_winres(libui_dir: &Path) -> io::Result<()> {
+    let windows_dir = libui_dir.join("windows");
+    winres::WindowsResource::new()
+        .set_manifest_file(&windows_dir.join("libui.manifest").display().to_string())
+        .compile()
+}
+
+fn link_kind() -> &'static str {
+    if cfg!(feature = "build") {
+        "static"
+    } else {
+        "dylib"
+    }
+}
+
 mod dep {
     use std::path::Path;
 
@@ -127,170 +143,211 @@ mod dep {
     impl rusync::progress::ProgressInfo for FakeProgressInfo {}
 }
 
-#[cfg(feature = "build")]
-mod libui {
-    use std::{io, path::Path};
+mod build {
+    use std::{env, fs, io, path::{Path, PathBuf}, process};
 
-    /// The error type returned by *libui* functions.
+    /// The error type returned by [`Backend`] functions.
     #[derive(Debug)]
     pub enum Error {
         /// Failed to setup *libui*.
-        SetupLibui(crate::meson::Error),
+        SetupLibui(PythonError),
         /// Failed to build Ninja.
-        BuildNinja(crate::ninja::Error),
-        /// Failed to build *libui* with Ninja.
-        BuildLibuiWithNinja(crate::ninja::Error),
-        /// Failed to build *libui* with Meson.
-        BuildLibuiWithMeson(crate::meson::Error),
-        /// Failed to rename "libui.a" to "ui.lib".
+        BuildNinja(PythonError),
+        /// Failed to compile *libui*.
+        CompileLibui(PythonError),
+        /// Failed to rename `libui.a` to `ui.lib`.
         ///
         /// This error *should* only occur when `$CARGO_CFG_TARGET_OS` is `windows`.
         RenameLibui(io::Error),
     }
 
-    /// Builds *libui*.
-    pub fn build(libui_dir: &Path, meson_dir: &Path, ninja_dir: &Path) -> Result<(), Error> {
-        crate::meson::setup_libui(meson_dir, libui_dir).map_err(Error::SetupLibui)?;
-        if build_cfg!(target_os = "linux") {
-            crate::ninja::build(ninja_dir).map_err(Error::BuildNinja)?;
-            crate::ninja::build_libui(ninja_dir, libui_dir).map_err(Error::BuildLibuiWithNinja)?;
-        } else {
-            crate::meson::build_libui(meson_dir, libui_dir).map_err(Error::BuildLibuiWithMeson)?;
-        }
-
-        // Meson unconditionally names the library "libui.a", which prevents MSVC's `link.exe` from
-        // finding it; we must manually rename it to "ui.lib".
-        if build_cfg!(target_os = "windows") {
-            let build_dir = libui_dir.join("build/meson-out");
-            std::fs::rename(build_dir.join("libui.a"), build_dir.join("ui.lib"))
-                .map_err(Error::RenameLibui)?;
-        }
-
-        Ok(())
-    }
-}
-
-#[cfg(feature = "build")]
-mod meson {
-    use std::{env, io, path::Path, process};
-
-    /// The error type returned by *meson* functions.
     #[derive(Debug)]
-    pub enum Error {
+    pub enum PythonError {
         /// Failed to run Python.
         RunPython(io::Error),
         /// The process run by Python failed.
         Python { out: process::Output },
     }
 
-    /// Prepares *libui* to be built.
-    pub fn setup_libui(meson_dir: &Path, libui_dir: &Path) -> Result<(), Error> {
-        let out = process::Command::new("python")
-            .arg(meson_dir.join("meson.py"))
-            .arg("setup")
-            .arg("--default-library=static")
-            .arg("--buildtype=release")
-            .arg(format!("--optimization={}", optimization()))
-            .arg(format!("--backend={}", backend()))
-            .arg("-Db_vscrt=from_buildtype")
-            .arg(libui_dir.join("build"))
-            .arg(libui_dir)
-            .output()
-            .map_err(Error::RunPython)?;
+    pub enum Backend {
+        Msvc,
+        Ninja,
+        Xcode,
+    }
+
+    impl Default for Backend {
+        fn default() -> Self {
+            if build_cfg!(feature = "build-with-msvc") {
+                Self::Msvc
+            } else if build_cfg!(feature = "build-with-xcode") {
+                Self::Xcode
+            // Ninja is last because it is the default option. This way, even if the user forgets to
+            // pass `--no-default-options` and both `build-with-ninja` and, e.g., `build-with-msvc`
+            // are enabled, only `build-with-msvc` will take effect, and the build backend will be
+            // MSVC.
+            } else if build_cfg!(feature = "build-with-ninja") {
+                Self::Ninja
+            } else {
+                panic!(
+                    "
+                    The `build` feature is enabled but no `build-with-*` feature is not enabled. \
+                    *libui-ng-sys* doesn't know which build backend to use. \
+                    "
+                );
+            }
+        }
+    }
+
+    impl Backend {
+        /// Builds *libui*.
+        pub fn build_libui(
+            self,
+            libui_dir: &Path,
+            meson_dir: &Path,
+            ninja_dir: &Path,
+        ) -> Result<(), Error> {
+            if Self::libui_path(libui_dir).exists() {
+                // We'll give the benefit of the doubt that this is actually a complete, working
+                // library.
+                return Ok(());
+            }
+
+            if let Self::Ninja = self {
+                // This must precede setting up *libui* as Meson requires Ninja even in the
+                // configuration phase.
+                Self::build_ninja(ninja_dir).map_err(Error::BuildNinja)?;
+            }
+
+            self.setup_libui(libui_dir, meson_dir, ninja_dir).map_err(Error::SetupLibui)?;
+            self.compile_libui(libui_dir, meson_dir, ninja_dir)
+                .map_err(Error::CompileLibui)?;
+            self.rename_libui(libui_dir).map_err(Error::RenameLibui)?;
 
-        if out.status.success() {
             Ok(())
-        } else {
-            Err(Error::Python { out })
         }
-    }
 
-    fn backend() -> &'static str {
-        if build_cfg!(target_os = "macos") {
-            todo!()
-        } else if build_cfg!(target_os = "linux") {
-            "ninja"
-        } else if build_cfg!(target_os = "windows") {
-            "vs"
-        } else {
-            unimplemented!("Unsupported target OS");
+        fn libui_path(libui_dir: &Path) -> PathBuf {
+            libui_dir.join("libui.a")
         }
-    }
 
-    fn is_debug() -> bool {
-        !matches!(env::var("DEBUG").as_deref(), Ok("0" | "false"))
-    }
-
-    fn optimization() -> String {
-        let level = env::var("OPT_LEVEL").expect("$OPT_LEVEL is unset");
-        match level.as_str() {
-            // Meson doesn't support "-Oz"; we'll try the next-closest option.
-            "z" => String::from("s"),
-            _ => level,
+        fn ninja_path(ninja_dir: &Path) -> PathBuf {
+            let ext = env::consts::EXE_EXTENSION;
+            ninja_dir.join("ninja").with_extension(ext)
         }
-    }
 
-    pub fn build_libui(meson_dir: &Path, libui_dir: &Path) -> Result<(), Error> {
-        let out = process::Command::new("python")
-            .arg(meson_dir.join("meson.py"))
-            .arg("compile")
-            .current_dir(libui_dir.join("build"))
-            .output()
-            .map_err(Error::RunPython)?;
+        fn run_python(
+            f: impl Fn(&mut process::Command),
+            ninja_dir: Option<&Path>,
+        ) -> Result<(), PythonError> {
+            let mut cmd = process::Command::new("python3");
+            f(&mut cmd);
+
+            if let Some(dir) = ninja_dir {
+                cmd.env("NINJA", Self::ninja_path(dir));
+            }
+
+            let out = cmd.output().map_err(PythonError::RunPython)?;
+            if out.status.success() {
+                Ok(())
+            } else {
+                Err(PythonError::Python { out })
+            }
+        }
+
+        /// Builds Ninja.
+        fn build_ninja(ninja_dir: &Path) -> Result<(), PythonError> {
+            if Self::ninja_path(ninja_dir).exists() {
+                // We'll give the benefit of the doubt that this is actually a complete, working
+                // binary.
+                return Ok(());
+            }
+
+            Self::run_python(
+                |cmd| {
+                    cmd
+                        .arg("configure.py")
+                        .arg("--bootstrap")
+                        .current_dir(ninja_dir);
+                },
+                None,
+            )
+        }
+
+        /// Prepares *libui* to be built.
+        fn setup_libui(
+            &self,
+            libui_dir: &Path,
+            meson_dir: &Path,
+            ninja_dir: &Path,
+        ) -> Result<(), PythonError> {
+            Self::run_python(
+                |cmd| {
+                    cmd
+                        .arg(meson_dir.join("meson.py"))
+                        .arg("setup")
+                        .arg("--default-library=static")
+                        .arg("--buildtype=release")
+                        .arg(format!("--optimization={}", Self::optimization_level()))
+                        .arg(format!("--backend={}", self.as_str()))
+                        // It's OK that this option is hardcoded (which is MSVC-specific) for all
+                        // backends; Meson will simply ignore it if MSVC isn't the selected backend.
+                        .arg("-Db_vscrt=from_buildtype")
+                        .arg(libui_dir.join("build"))
+                        .arg(libui_dir);
+                },
+                Some(ninja_dir),
+            )
+        }
+
+        // This may be used at some point.
+        #[allow(dead_code)]
+        fn is_debug() -> bool {
+            !matches!(env::var("DEBUG").as_deref(), Ok("0" | "false"))
+        }
+
+        fn optimization_level() -> String {
+            let level = env::var("OPT_LEVEL").expect("$OPT_LEVEL is unset");
+            match level.as_str() {
+                // Meson doesn't support "-Oz"; we'll try the next-closest option.
+                "z" => String::from("s"),
+                _ => level,
+            }
+        }
+
+        fn as_str(&self) -> &'static str {
+            match self {
+                Self::Msvc => "vs",
+                Self::Ninja => "ninja",
+                Self::Xcode => "xcode",
+            }
+        }
+
+        fn compile_libui(
+            &self,
+            libui_dir: &Path,
+            meson_dir: &Path,
+            ninja_dir: &Path,
+        ) -> Result<(), PythonError> {
+            Self::run_python(
+                |cmd| {
+                    cmd
+                        .arg(meson_dir.join("meson.py"))
+                        .arg("compile")
+                        .arg(format!("-C={}", libui_dir.join("build").display()));
+                },
+                Some(ninja_dir),
+            )
+        }
+
+        fn rename_libui(&self, libui_dir: &Path) -> Result<(), io::Error> {
+            // Meson unconditionally names the library "libui.a", which prevents MSVC's `link.exe`
+            // from finding it; we must manually rename it to "ui.lib".
+            if let Self::Msvc = self {
+                let build_dir = libui_dir.join("build/meson-out");
+                fs::rename(Self::libui_path(libui_dir), build_dir.join("ui.lib"))?;
+            }
 
-        if out.status.success() {
             Ok(())
-        } else {
-            Err(Error::Python { out })
-        }
-    }
-}
-
-#[cfg(feature = "build")]
-mod ninja {
-    use std::{io, path::Path, process};
-
-    /// The error type returned by *ninja* functions.
-    #[derive(Debug)]
-    pub enum Error {
-        /// Failed to run Python.
-        RunPython(io::Error),
-        /// The process run by Python failed.
-        Python { out: process::Output },
-    }
-
-    /// Builds Ninja.
-    pub fn build(ninja_dir: &Path) -> Result<(), Error> {
-        if ninja_dir.join("ninja").exists() {
-            return Ok(());
-        }
-
-        let out = std::process::Command::new("python3")
-            .arg("configure.py")
-            .arg("--bootstrap")
-            .current_dir(ninja_dir)
-            .output()
-            .map_err(Error::RunPython)?;
-
-        if out.status.success() {
-            Ok(())
-        } else {
-            Err(Error::Python { out })
-        }
-    }
-
-    /// Builds *libui* with Ninja after configuration with Meson.
-    pub fn build_libui(ninja_dir: &Path, libui_dir: &Path) -> Result<(), Error> {
-        let out = std::process::Command::new(ninja_dir.join("ninja"))
-            .args(["-C", "build"])
-            .current_dir(libui_dir)
-            .output()
-            .map_err(Error::RunPython)?;
-
-        if out.status.success() {
-            Ok(())
-        } else {
-            Err(Error::Python { out })
         }
     }
 }
@@ -399,7 +456,8 @@ mod bindings {
                 .parse_callbacks(Box::new(bindgen::CargoCallbacks))
                 .allowlist_function(LIBUI_REGEX)
                 .allowlist_type(LIBUI_REGEX)
-                .allowlist_var(LIBUI_REGEX);
+                .allowlist_var(LIBUI_REGEX)
+                .blocklist_item("_bindgen.*");
 
             // Note: Virtually every wrapper except that for "ui.h" should blocklist "ui.h".
             if self.blocklists_main {
@@ -408,6 +466,7 @@ mod bindings {
 
             builder
                 .clang_args(ClangArgs::new().as_args())
+                .layout_tests(false)
                 .generate()
                 .map_err(|_| Error::Generate)?
                 .write_to_file(out_dir.join(format!("{}.rs", self.filename)))
@@ -462,9 +521,9 @@ mod bindings {
     impl ClangArgs {
         fn new() -> Self {
             if build_cfg!(target_os = "macos") {
-                Self::new_darwin()
+                Self::new_macos()
             } else if build_cfg!(target_os = "linux") {
-                Self::new_unix()
+                Self::new_linux()
             } else if build_cfg!(target_os = "windows") {
                 Self::new_windows()
             } else {
@@ -472,15 +531,14 @@ mod bindings {
             }
         }
 
-        fn new_darwin() -> Self {
-            // TODO
+        fn new_macos() -> Self {
             Self {
                 defines: Vec::new(),
                 include_paths: Vec::new(),
             }
         }
 
-        fn new_unix() -> Self {
+        fn new_linux() -> Self {
             let gtk = pkg_config::Config::new()
                 .atleast_version("3.10.0")
                 .print_system_cflags(true)
@@ -499,7 +557,7 @@ mod bindings {
             let include_paths = gtk
                 .include_paths
                 .into_iter()
-                .map(|path| format!("{}", path.display()))
+                .map(|path| path.display().to_string())
                 .collect();
 
             Self {
